﻿using System.Collections.Generic;
using System.Linq;
using QuickGraph;
using QuickGraph.Algorithms;
using QuickGraph.Algorithms.Search;
using QuickGraph.Algorithms.ShortestPath;
using StudyCrafter.Assistant.Representation;
using StudyCrafter.Assistant.StudyCrafterScenario;

namespace StudyCrafter.Assistant.StudyCrafterLoading
{
    public class InteractionMapPostProcAlgorithm : AlgorithmBase<InteractionMap>
    {
		// TODO:
		//
		// - Possibly add creation of indirect edges.

		private IEdgeIDGenerator<InteractionNode, InteractionEdge> edgeIDGen;

		public InteractionMapPostProcAlgorithm(InteractionMap visitedGraph, IEdgeIDGenerator<InteractionNode, InteractionEdge> edgeIDGen) : base(visitedGraph)
		{
			this.edgeIDGen = edgeIDGen;
		}

        public InteractionMapPostProcAlgorithm(InteractionMap visitedGraph) : base(visitedGraph)
        {
			edgeIDGen = new SequentialEdgeIDGenerator<InteractionNode, InteractionEdge>();
        }

        protected override void InternalCompute()
        {
            // Fix any dead-ends that should be escapable through events.
            FixUpDeadEnds();

            AssociateInteractionUnits();

            FixUpNodeTypes();

        }

        private void FixUpDeadEnds()
        {
            IEnumerable<InteractionNode> endNodes = VisitedGraph.FindEndNodes();

            IEnumerable<InteractionNode> events = VisitedGraph.FindEventBasedInteractionNodes();

            DijkstraShortestPathAlgorithm<InteractionNode, InteractionEdge> distanceFromStart = new DijkstraShortestPathAlgorithm<InteractionNode, InteractionEdge>(VisitedGraph.Graph, InteractionMapExtensions.AllEdgesAvailable);
            distanceFromStart.Compute(VisitedGraph.GetStart());

            IEnumerable<InteractionNode> nodesSortedByInverseStartDistance = VisitedGraph.Graph.Vertices.OrderByDescending((interactionNode) =>
            {
                double distance = double.MaxValue;
                if (!distanceFromStart.TryGetDistance(interactionNode, out distance))
                {
                    return double.MaxValue;
                }
                return distance;
            });

            foreach (InteractionNode node in nodesSortedByInverseStartDistance)
            {
                if (!VisitedGraph.HasPathToAtLeastOne(node, endNodes))
                {
                    // This is a dead end. Attach it to all events.
                    foreach (InteractionNode eventNode in events)
                    {
						InteractionEdge indirectEdge = new InteractionEdge(InteractionEdge.Type.Indirect, node, eventNode);
						indirectEdge.SetID(GenerateEdgeID(indirectEdge));
                        VisitedGraph.Graph.AddEdge(indirectEdge);
                    }
                }
            }
        }

        private void AssociateInteractionUnits()
        {
            foreach (InteractionNode node in VisitedGraph.Graph.Vertices)
            {
                if (node.NodeType == InteractionNode.Type.InteractionPoint)
                {
                    // Associate it with its own unit ID, to start.
                    node.AssociateWithUnit(node.ID);

                    DijkstraShortestPathAlgorithm<InteractionNode, InteractionEdge> pathAlg = new DijkstraShortestPathAlgorithm<InteractionNode, InteractionEdge>(VisitedGraph.Graph, InteractionMapExtensions.StopAtInteractionPoints);
                    pathAlg.TreeEdge += (edge) => EdgeReachableFromInteractionPoint(edge, node.ID);
                    pathAlg.Compute(node);
                }
            }
        }

        private void FixUpNodeTypes()
        {
            DepthFirstSearchAlgorithm<InteractionNode, InteractionEdge> dfs =
                new DepthFirstSearchAlgorithm<InteractionNode, InteractionEdge>(this, VisitedGraph.Graph, new Dictionary<InteractionNode, GraphColor>(VisitedGraph.Graph.VertexCount));
            dfs.DiscoverVertex += HandleVertex;

            dfs.Compute();
        }

        private void EdgeReachableFromInteractionPoint(InteractionEdge edge, string unitID)
        {
            edge.Source.AssociateWithUnit(unitID);
            edge.Target.AssociateWithUnit(unitID);
        }

        private void HandleVertex(InteractionNode node)
        {
            FixUpNodeType(node);
        }

        private void FixUpNodeType(InteractionNode node)
        {
            if (node.NodeType == InteractionNode.Type.Unknown)
            {
                node.RecomputeType();
            }
        }

		private int GenerateEdgeID(InteractionEdge edge)
		{
			return edgeIDGen.GenerateID(edge);
		}
    }
}
