﻿using System;
using System.Collections.Generic;
using Newtonsoft.Json;
using StudyCrafter.Assistant.Loading;
using StudyCrafter.Assistant.Representation;
using StudyCrafter.Assistant.StudyCrafterScenario;
using UnityEngine;

namespace StudyCrafter.Assistant.StudyCrafterLoading
{
	[RequireComponent(typeof(StreamingAssetsProjectLoader))]
	public class ProjectDirectoryLoader : MonoBehaviour
	{
		private string BasePath
		{
			get
			{
				return Application.streamingAssetsPath;
			}
		}

		[SerializeField]
		private string projectsPath = "NewProjectSystem";

		public string ProjectsFolder
		{
			get
			{
				return projectsPath;
			}
			set
			{
				projectsPath = value;
			}
		}

		[SerializeField]
		private string scriptNodeDefsPath = "Definitions/ScriptNodeDefs";

		private IProjectLoader projectLoader;

		private IGraphStore graphStore = new GraphStore();

		private IGraphIDGenerator graphIDGenerator = new SequentialIDGenerator();

		private ScriptNodeDefCollection scriptNodeDefs = null;

		private Dictionary<string, Scenario> scenarios = new Dictionary<string, Scenario>();

		private Dictionary<string, ProjectLoadStatus> loadStatuses = new Dictionary<string, ProjectLoadStatus>();

		private ScenarioLoader scenarioLoader = new ScenarioLoader();

		public delegate void CompleteEvent();

		public CompleteEvent Complete;

		public delegate void ErrorEvent(string error);

		public event ErrorEvent ErrorHandler;

		private bool isIdle = true;

		public IGraphStore GraphStore
		{
			get
			{
				return graphStore;
			}
		}


		public IDictionary<string, Scenario> Scenarios
		{
			get
			{
				return scenarios;
			}
		}

		public ProjectDirectoryLoader()
		{
		}

		public void Awake()
		{
			projectLoader = GetComponent<StreamingAssetsProjectLoader>();
		}

		public void Start()
		{
			Debug.Log("Loading script node definitions...");

			LoadScriptNodeDefs();
		}

		public void Update()
		{
			if (isIdle)
			{
				return;
			}

			if (loadStatuses.Count == 0)
			{
				return;
			}

			bool done = true;
			foreach (KeyValuePair<string, ProjectLoadStatus> loadStatus in loadStatuses)
			{
				if (loadStatus.Value.Status == LoadStatus.Loaded)
				{
					Debug.Log("Loaded: " + loadStatus.Key);
					OnLoaded(loadStatus.Key);
					loadStatus.Value.UpdateStatus(LoadStatus.None);
				}
				else if (loadStatus.Value.Status == LoadStatus.Error)
				{
					Debug.LogError("Loading Error: " + loadStatus.Value.ErrorText);
					OnError(loadStatus.Value.ErrorText);
					break;
				}

				if (loadStatus.Value.Status == LoadStatus.InProgress)
				{
					done = false;
				}
			}

			if (done)
			{
				OnComplete();
			}
		}

		public void LoadProjects()
		{
			if (!isIdle)
			{
				Debug.LogError("Can't restart - analysis in progress.");
				return;
			}

			Clear();

			isIdle = false;

			Debug.Log("Loading projects...");
			string path = System.IO.Path.Combine(BasePath, projectsPath);
			try
			{
				IEnumerable<string> projects = System.IO.Directory.GetDirectories(path);
				foreach (string projectFullPath in projects)
				{
					string project = System.IO.Path.GetFileName(projectFullPath);
					Debug.Log("Loading: " + project);
					ProjectLoadStatus status = projectLoader.LoadProject(project);
					loadStatuses.Add(project, status);
				}
			}
			catch (Exception e)
			{
				Debug.LogException(e);
				OnError(e.Message);
			}
		}

		private void LoadScriptNodeDefs()
		{
			// FIXME: Should this be loaded from streaming assets instead?
			TextAsset scriptNodeDefsFile = Resources.Load(scriptNodeDefsPath) as TextAsset;

			scriptNodeDefs = JsonConvert.DeserializeObject<ScriptNodeDefCollection>(scriptNodeDefsFile.text);
		}

		private void OnComplete()
		{
			Debug.Log("All projects loaded.");
			isIdle = true;
			if (Complete != null)
			{
				Complete();
			}
		}

		private void OnError(string message)
		{
			Debug.LogError(message);
			isIdle = true;
			if (ErrorHandler != null)
			{
				ErrorHandler(message);
			}
		}

		public bool IsComplete()
		{
			return isIdle;
		}

		public bool IsLoaded(string project)
		{
			ProjectLoadStatus loadStatus;
			if (!loadStatuses.TryGetValue(project, out loadStatus))
			{
				return false;
			}
			return loadStatus.Status == LoadStatus.None;
		}

		private void OnLoaded(string project)
		{
			JsonScenarioStrings loadedProject = projectLoader.GetLoadedProject(project);
			StudyCrafterLoadHelpers loadHelpers = new StudyCrafterLoadHelpers(graphStore, graphIDGenerator, scriptNodeDefs);
			try
			{
				Scenario scenario = scenarioLoader.Load(loadedProject, loadHelpers);
				scenarios.Add(scenario.ID, scenario);
			}
			catch (Exception e)
			{
				OnError(e.Message);
				Debug.LogException(e);
			}

			Debug.Log("Loaded:" + project);
		}

		public void Clear()
		{
			isIdle = true;
			loadStatuses = new Dictionary<string, ProjectLoadStatus>();
			graphStore = new GraphStore();
			graphIDGenerator = new SequentialIDGenerator();
			scenarios = new Dictionary<string, Scenario>();
		}
	}
}
