﻿using System;
using System.Collections.Generic;
using QuickGraph;
using StudyCrafter.Assistant.Representation;
using StudyCrafter.Assistant.StudyCrafterScenario;
using UnityEngine;

namespace StudyCrafter.Assistant.StudyCrafterLoading
{
    public static class ScriptGraphJsonLoadingExtensions
    {
        public static ScriptGraph FromSerializationData(this ScriptGraph scriptGraph, ScriptSerializationData scriptData, ScriptNodeDefCollection nodeDefs)
        {
            return scriptGraph.CreateVertices(scriptData, nodeDefs)
                              .CreateEdges();
        }

        public static ScriptGraph CreateVertices(this ScriptGraph scriptGraph, ScriptSerializationData scriptData, ScriptNodeDefCollection nodeDefs)
        {
            foreach (KeyValuePair<string, ScriptNodeSerializationData> scriptEntry in scriptData)
            {
                ScriptNode node = new ScriptNode(scriptEntry.Key)
                    .FromSerializationData(scriptEntry.Value, nodeDefs);

                scriptGraph.Graph.AddVertex(node);
            }
            return scriptGraph;
        }

        public static ScriptGraph CreateEdges(this ScriptGraph scriptGraph)
        {
            foreach (ScriptNode vertex in scriptGraph.Graph.Vertices)
            {
                scriptGraph.CreateEdgesForVertex(vertex);
            }

            return scriptGraph;
        }

        public static ScriptGraph CreateEdgesForVertex(this ScriptGraph scriptGraph, ScriptNode vertex)
        {
            // Create edges to any possible next nodes the script can transition to from this one.
            foreach (string nextNode in vertex.GetNextNodes())
            {
                ScriptNode otherVertex;
                if (!scriptGraph.TryGetVertex(new ScriptNode(nextNode), out otherVertex))
                {
                    Debug.LogWarning("Next Script Node not found: " + nextNode);
                    continue;
                }
                scriptGraph.Graph.AddEdge(new Edge<ScriptNode>(vertex, otherVertex));
            }

            return scriptGraph;
        }

        public static IEnumerable<string> GetNextNodes(this ScriptNode node)
        {
            switch (node.Type)
            {
                case ScriptNodeTypes.Choice:
                    return node.GetNextNodesChoice();
                case ScriptNodeTypes.Branch:
                    return node.GetNextNodesBranch();
                default:
                    return node.GetNextNodesStandard();
            }
        }

        public static IEnumerable<string> GetNextNodesChoice(this ScriptNode node)
        {
            // Open-response nodes actually use a standard "next" property.
            if (node.GetProperty<StringProperty>("choiceType").Val.Equals("open_response"))
            {
                return node.GetNextNodesStandard();
            }

            HashSet<string> nextNodes = new HashSet<string>();

            ChoicesProperty choices;
            if (!node.TryGetProperty("choices", out choices))
            {
                return nextNodes;
            }

            foreach (ChoicesProperty.Choice choice in choices.Choices)
            {
                if (choice.Index.IntValue < 0)
                {
                    continue;
                }

                nextNodes.Add(choice.Index.Val);
            }

            return nextNodes;
        }

        public static IEnumerable<string> GetNextNodesBranch(this ScriptNode node)
        {
            HashSet<string> nextNodes = new HashSet<string>();

            ConditionalProperty conditions;
            if (!node.TryGetProperty("choices", out conditions))
            {
                return nextNodes;
            }

            foreach (ConditionalProperty.Branch branch in conditions.Branches)
            {
                if (branch.Index.IntValue < 0)
                {
                    continue;
                }

                nextNodes.Add(branch.Index.Val);
            }

            return nextNodes;
        }

        public static IEnumerable<string> GetNextNodesStandard(this ScriptNode node)
        {
            HashSet<string> nextNodes = new HashSet<string>();

            // Check if we have a valid next node.
            IntProperty nextNodeProperty;
            if (node.TryGetProperty("next", out nextNodeProperty))
            {
                if (nextNodeProperty.IntValue >= 0)
                {
                    nextNodes.Add(nextNodeProperty.Val);
                }
            }

            return nextNodes;
        }
    }
}